use bincode::{self, options, Options};
use std::{net::{UdpSocket, SocketAddr}, io::Read};
use serde::{Serialize, Deserialize};
use sha3::{Digest, Sha3_512};
use sha2::Sha512;
use regex::Regex;

mod big_array;
use big_array::BigArray;

const MAX_FRAME_PAYLOAD:u16=508;
const MAX_FRAME_PAYLOAD_U:usize=MAX_FRAME_PAYLOAD as usize;
const HEADER_SIZE:u16 = 20;
const MAX_PAYLOAD:u16 = MAX_FRAME_PAYLOAD - HEADER_SIZE;
const MAX_PAYLOAD_U:usize = MAX_PAYLOAD as usize;

#[derive(Serialize, Deserialize)]
struct PacketInformation {
    packet_numbers: u32, //4 bytes
    last_packet_size: u16, //2 bytes
    response_filename_checksum: u64, //8 bytes
} //14 bytes

#[derive(Serialize, Deserialize)]
struct Packet {
    packet_number: u32, //4 bytes
    payload_hash: [u8; 16], //16 bytes
    #[serde(with = "BigArray")]
    payload: [u8; MAX_PAYLOAD_U], //488 bytes
} //512 bytes

#[derive(Serialize, Deserialize)]
struct StrPacket {
    #[serde(with = "BigArray")]
    payload: [u8; MAX_PAYLOAD_U]
}

fn main() {
    let port = "1337";
    let timeout = 100;
    let local_addr: SocketAddr = ("0.0.0.0:".to_string()+port).parse().expect("Failed to parse address");
    let socket = UdpSocket::bind(local_addr).expect("Failed to bind socket");

    socket.set_read_timeout(Some(std::time::Duration::from_millis(timeout))).expect("set_read_timeout call failed");

    println!("UDP Server up and running on port {}", local_addr.port());

    let options = options().with_big_endian().allow_trailing_bytes().with_fixint_encoding();

    let hash_request_regex = Regex::new(r"[a-zA-Z0-9.-_ ]+:").unwrap();

    let missing_packet_request_regex = Regex::new(r"([a-zA-Z0-9.-_ ]+)(/[0-9]+)+").unwrap();

    loop {
        let mut buffer = [0u8; MAX_FRAME_PAYLOAD_U];
        let res = socket.recv_from(&mut buffer);
        if let Ok((_received_bytes, remote_addr)) = res {
            let filled_buffer = &buffer;//[..received_bytes];

            let request_packet = bincode::deserialize::<StrPacket>(filled_buffer).unwrap();
            let mut request: String = request_packet.payload.iter().map(|&c| c as char).collect();
            request.retain(|c| c != '\0');

            println!("Received request: {}", request);
            let req = request.as_str();

            if hash_request_regex.is_match(req) {
                println!("Received hash request");
                continue;
            }

            if missing_packet_request_regex.is_match(req) {
                println!("Received missing packet request");
                continue;
            }

            println!("Received file request");

            let file_result = std::fs::File::open("./files/".to_string() + req);

            if file_result.is_err() {
                println!("{}",file_result.err().unwrap().to_string());
                println!("File not found");
                continue;
            }

            let mut file = file_result.unwrap();

            let mut file_buffer = [0u8; MAX_PAYLOAD_U];

            let mut packet_number = 0u32;

            let file_length = file.metadata().unwrap().len();
            println!("file length: {}", file_length);

            let packet_numbers = f64::ceil(file_length as f64/MAX_PAYLOAD as f64) as u32;

            let packet_information = PacketInformation {
                packet_numbers,
                last_packet_size: (file_length%MAX_PAYLOAD as u64) as u16,
                response_filename_checksum: request_packet.payload.iter().fold(0u64, |acc, &x| acc.wrapping_add(x as u64)),
            };

            socket.send_to(options.serialize(&packet_information).unwrap().as_slice(), remote_addr).expect("Failed to send packet information");

            let mut sha512_hasher = Sha512::new();

            //read file chunk by chunk

            while let Ok(bytes_read) = file.read(&mut file_buffer) {
                if bytes_read == 0 {
                    break;
                }

                let mut hasher = Sha3_512::new();
                hasher.update(&file_buffer[..bytes_read]);
                let result = hasher.finalize();
                let mut packet = Packet {
                    packet_number,
                    payload_hash: [0u8; 16],
                    payload: [0u8; MAX_PAYLOAD_U],
                };
                packet.payload_hash.copy_from_slice(&result[..16]);
                packet.payload[..bytes_read].copy_from_slice(&file_buffer[..bytes_read]);
                let packet_bytes = options.serialize(&packet).unwrap();
                socket.send_to(&packet_bytes, remote_addr).expect("Failed to send packet");
                packet_number += 1;

                sha512_hasher.update(&file_buffer[..bytes_read]);
            }

            let result = sha512_hasher.finalize();
            //add 1 byte of padding to result
            let mut result_bytes = [0u8; 65];
            result_bytes[..64].copy_from_slice(&result[..64]);

            socket.send_to(&result_bytes, remote_addr).expect("Failed to send file hash");
        }
    }
}
